// EllipticButton.cpp: implementation file
//

#include "stdafx.h"
#include "EllipticButton.h"
#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


#define	PI	3.141592
/////////////////////////////////////////////////////////////////////////////
// CEllipticButton

CEllipticButton::CEllipticButton()
{
	m_selectedState = false;
	m_defPushButton = false;
}

CEllipticButton::~CEllipticButton()
{
}

BEGIN_MESSAGE_MAP(CEllipticButton, CButton)
	//{{AFX_MSG_MAP(CEllipticButton)
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CEllipticButton message handlers

/*****************************************************************************************/
/* this handler must be overwrite if the property owner draw is selected                 */
/* there are all stuff for the state and the style of the elliptic button to each time   */
/*****************************************************************************************/
void CEllipticButton::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct) 
{
	ASSERT(lpDrawItemStruct != NULL);
	
	// get any states and styles
	CRect rect = lpDrawItemStruct->rcItem;
	UINT state = lpDrawItemStruct->itemState;
	DWORD nStyle = GetStyle();

	// if the bitmap missing the bitmap style must remove
	if((nStyle & BS_BITMAP) && !GetBitmap()) {
		nStyle = nStyle ^ BS_BITMAP;
	}
 
	// get the color of the standard button
	COLORREF sysColor = GetSysColor(COLOR_BTNFACE);
	// create any pens
	CPen penWhite(PS_SOLID, 0, RGB(255, 255, 255));
	CPen penBlack(PS_SOLID, 0, RGB(0, 0, 0));

	CPen penTrans;
	LOGPEN logPen;
	logPen.lopnStyle = PS_SOLID;
	logPen.lopnColor = sysColor;
	logPen.lopnWidth = CPoint(1,1);
	penTrans.CreatePenIndirect(&logPen);

	// set the region of the elliptic button
	CRgn rgn;
	SetWindowRgn(NULL, FALSE);
	rgn.CreateEllipticRgnIndirect(rect);
	SetWindowRgn(rgn, FALSE);

	// memory context for the drawing operations
	CDC* pDC   = GetDC();
	
	// set the prefered brush
	CBrush brush(sysColor);
	pDC->SelectObject(&brush);

	// styles
	// style bitmap
	// the bitmap would be not drawing in the whole range
	CRect rectBitmap = rect;
	rectBitmap.top++;
	rectBitmap.left++;

	// seperate memory context for the bitmap
	// (all what displayed on the button must be save in the bitmap, inclusive all text)
	CDC dcBmpButton;
	dcBmpButton.CreateCompatibleDC(pDC);

	m_bitmap.CreateCompatibleBitmap(pDC, rect.Width(), rect.Height());
	dcBmpButton.SelectObject(&m_bitmap);
	dcBmpButton.FillRect(&rect, &brush);	// fills the bitmap with the original button bachground

	pDC->SelectObject(&brush);

	CRgn rgnComb;
	rgnComb.CreateEllipticRgnIndirect(rectBitmap);
	dcBmpButton.SelectObject(&rgnComb);

	BOOL bBitmap = (nStyle & BS_BITMAP);

	int posx = 0;
	int posy = 0;

	BITMAP bmp = BITMAP();

	CDC dcBmpOriginal;
	CBitmap bitmapOriginal;		// the original bitmap

	// Bitmap in dcBmp holen
	HBITMAP hBitmap = GetBitmap();
	bitmapOriginal.Attach(hBitmap);

	if(bBitmap) {

		bitmapOriginal.GetBitmap(&bmp);

		// a memory context for the original bitmap
		dcBmpOriginal.CreateCompatibleDC(pDC);
		dcBmpOriginal.SelectObject(&bitmapOriginal);

		// position of the original bitmap on the button bitmap
		int posxi = 0, posyi = 0;
		posx = (rectBitmap.Width() - bmp.bmWidth)/2 + 1;
		posy = (rectBitmap.Height() - bmp.bmHeight)/2 + 1;

		if(posx < 0) posxi = posx - 2;
		if(posy < 0) posyi = posy - 2;

		// the original bitmap will be copy in the center of the button bitmap
		dcBmpButton.BitBlt(posx, posy, bmp.bmWidth, bmp.bmHeight, &dcBmpOriginal, 0, 0, SRCCOPY);
	}

	// also the text will be display in the bitmap
	// if the user don't like text on the button, he must remove the window title in 
	// the resource editor

	CFont* pFont = GetParent()->GetFont();		// Get the font of the parent window
	dcBmpButton.SelectObject(pFont);
	CString text;								// get the title (text) of the button
	GetWindowText(text);						// if no text is desired, make title empty
	CSize Extent = dcBmpButton.GetTextExtent(text);	// get the size of the text
	CRect rectText;

	rectText.top = 0;
	rectText.bottom = 1000;
	rectText.left = 0;
	rectText.right = rect.Width();

	UINT format = DT_CENTER | DT_VCENTER;

	rectText.left += 8;
	rectText.right -= 8;
	if(((nStyle & BS_CENTER) == BS_RIGHT) || ((nStyle & BS_CENTER) == WS_EX_RIGHT)) {
		format = DT_RIGHT;
	}
	else if((nStyle & BS_CENTER) == BS_LEFT) {
		format = DT_LEFT;
	}
	
	format |= DT_PATH_ELLIPSIS;
	if(nStyle & BS_MULTILINE) {
		format |= DT_WORDBREAK;
	}

	// only to check the height of text in a separate CDC
	CDC dc;
	dc.CreateCompatibleDC(&dcBmpButton);
	dc.SelectObject(pFont);
	int height = dc.DrawText(text, rectText, format);
	dc.DeleteDC();
	// end check height

	rectText.top = rect.Height()/2 - height/2;
	rectText.bottom = rect.Height()/2 + height/2 + height%2;

	if((nStyle & BS_VCENTER) == BS_TOP) {
		rectText.top = 8;
		rectText.bottom = rect.Height() - 8;
	}
	else if((nStyle & BS_VCENTER) == BS_BOTTOM) {
		rectText.top = rect.Height() - 8 - height;
		rectText.bottom = rect.Height() - 8;
	}
	dcBmpButton.SetBkMode(TRANSPARENT);
	if((state & ODS_DISABLED)) {
		dcBmpButton.SetBkColor(sysColor);
		dcBmpButton.SetTextColor(GetSysColor(COLOR_BTNHIGHLIGHT));
		rectText.OffsetRect(1, 1);
		dcBmpButton.DrawText(text, rectText, format);
		rectText.OffsetRect(-1, -1);
		dcBmpButton.SetTextColor(GetSysColor(COLOR_GRAYTEXT));
	}
	else {
		dcBmpButton.SetTextColor(GetSysColor(COLOR_BTNTEXT));
	}

	dcBmpButton.DrawText(text, rectText, format);

	// depends from the state the bitmap will be copy to the button
	// only 3 states must be consider (selected/ disable/ normal)
	if(m_selectedState) {		// maybe the selection state must be corrected, if the user moves the mouse
		state |= 0x00000001;	// selected
	}
	else {
		state &= 0xFFFFFFFE;	// not selected
	}

	if(state & ODS_SELECTED) {
		pDC->BitBlt(0, 0, rect.Width(), rect.Height(), &dcBmpButton, -1, -1, SRCCOPY);
	}
	else {
		if(state & ODS_DISABLED) {

			COLORREF disableColor = GetSysColor(COLOR_GRAYTEXT);

			CBrush disableBrush(disableColor);
			CBrush sysBrush(sysColor);
			CRect rectBmp(0, 0, bmp.bmWidth, bmp.bmHeight);

			// a monochrome CDC and a monochrome bitmap
			CDC dcMono;
			dcMono.CreateCompatibleDC(&dcBmpButton);
			CBitmap bitmapMono;
			bitmapMono.CreateBitmap(bmp.bmWidth, bmp.bmHeight, 1, 1, NULL);
			dcMono.SelectObject(bitmapMono);

			// and a CDC and bitmap for help operations
			CDC dcHelp;
			dcHelp.CreateCompatibleDC(&dcBmpButton);
			CBitmap bitmapHelp;
			bitmapHelp.CreateCompatibleBitmap(&dcBmpButton, bmp.bmWidth, bmp.bmHeight);

			dcHelp.SelectObject(&bitmapHelp);
			dcHelp.FillRect(rectBmp, &disableBrush);

			// and a CDC and bitmap for help operations
			CDC dcHelp1;
			dcHelp1.CreateCompatibleDC(&dcBmpButton);
			CBitmap bitmapHelp1;
			bitmapHelp1.CreateCompatibleBitmap(&dcBmpButton, bmp.bmWidth, bmp.bmHeight);

			dcHelp1.SelectObject(&bitmapHelp1);
			dcHelp1.FillRect(rectBmp, &sysBrush);
			
			dcMono.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcMono, 0, 0, NOTSRCCOPY);
			// the original bitmap will be copy in the center of the button bitmap
			dcMono.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcBmpOriginal, 0, 0, SRCCOPY);
			dcMono.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcMono, 0, 0, NOTSRCCOPY);
			dcHelp.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcHelp, 0, 0, NOTSRCCOPY);
			dcHelp.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcMono, 0, 0, SRCAND);
			dcHelp.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcHelp, 0, 0, NOTSRCCOPY);

			dcMono.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcMono, 0, 0, NOTSRCCOPY);
			dcHelp1.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcHelp1, 0, 0, NOTSRCCOPY);
			dcHelp1.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcMono, 0, 0, SRCAND);
			dcHelp1.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcHelp1, 0, 0, NOTSRCCOPY);

			dcHelp.BitBlt(0, 0, bmp.bmWidth, bmp.bmHeight, &dcHelp1, 0, 0, SRCAND);

			dcBmpButton.BitBlt(posx, posy, bmp.bmWidth, bmp.bmHeight, &dcHelp, 0, 0, SRCCOPY);

			// extra text handling


			bitmapHelp.DeleteObject();
			bitmapHelp1.DeleteObject();
			bitmapMono.DeleteObject();
			dcMono.DeleteDC();
			dcHelp.DeleteDC();
			dcHelp1.DeleteDC();
		}
		pDC->BitBlt(0, 0, rect.Width(), rect.Height(), &dcBmpButton, 0, 0, SRCCOPY);
	}

	rgnComb.Detach();
	rgnComb.DeleteObject();

	bitmapOriginal.Detach();
	bitmapOriginal.DeleteObject();

	// style flat
	if(nStyle & BS_FLAT) {
		// transparent background
		CBrush brush;
		LOGBRUSH logBrush;
		logBrush.lbStyle = BS_NULL;		// transparent brush
		brush.CreateBrushIndirect(&logBrush);
		CBrush* oldBrush = NULL;
		oldBrush = pDC->SelectObject(&brush);
		
		pDC->SelectObject(penWhite);	// white ellipse
		pDC->Ellipse(2, 2, rect.Width() - 2, rect.Height() - 2);
		pDC->SelectObject(penBlack);	// black ellipse
		pDC->Ellipse(1, 1, rect.Width() - 1, rect.Height() - 1);
	}
	else {	// style normal
		CPoint center = rect.CenterPoint();
		CPoint right = center;
		CPoint left = center;
		for(int i = 0; i < 360; i++) {
			COLORREF color = RGB(255 - i*255/360, 255 - i*255/360, 255 - i*255/360);
			right.x = center.x + (int)(sin(PI*i/360 - PI/4)*center.x - 1);
			right.y = center.y - (int)(cos(PI*i/360 - PI/4)*center.y);
			pDC->SetPixel(right, color);
			right.x = center.x + (int)(sin(PI*i/360 - PI/4 + PI/160)*center.x - 1);
			right.y = center.y - (int)(cos(PI*i/360 - PI/4 + PI/160)*center.y);
			pDC->SetPixel(right, color);
			right.x = center.x + (int)(sin(PI*i/360 - PI/4 + PI/80)*center.x - 1);
			right.y = center.y - (int)(cos(PI*i/360 - PI/4 + PI/80)*center.y);
			pDC->SetPixel(right, color);
			right.x = center.x + (int)(sin(PI*i/360 - PI/4 + 3*PI/160)*center.x - 1);
			right.y = center.y - (int)(cos(PI*i/360 - PI/4 + 3*PI/160)*center.y);
			pDC->SetPixel(right, color);

			color = RGB(i*255/360, i*255/360, i*255/360);
			right.x = center.x + (int)(1 - sin(PI*i/360 - PI/4)*center.x - 1);
			right.y = center.y - (int)(1 - cos(PI*i/360 - PI/4)*center.y);
			pDC->SetPixel(right, color);
			right.x = center.x + (int)(1 - sin(PI*i/360 - PI/4 - PI/160)*center.x - 1);
			right.y = center.y - (int)(1 - cos(PI*i/360 - PI/4 - PI/160)*center.y);
			pDC->SetPixel(right, color);
			right.x = center.x + (int)(1 - sin(PI*i/360 - PI/4 + PI/80)*center.x - 1);
			right.y = center.y - (int)(1 - cos(PI*i/360 - PI/4 + PI/80)*center.y);
			pDC->SetPixel(right, color);
			right.x = center.x + (int)(1 - sin(PI*i/360 - PI/4 + 3*PI/160)*center.x - 1);
			right.y = center.y - (int)(1 - cos(PI*i/360 - PI/4 + 3*PI/160)*center.y);
			pDC->SetPixel(right, color);
		}
	}
	
	rgnComb.DeleteObject();
	ReleaseDC(&dcBmpOriginal);
	dcBmpOriginal.DeleteDC();
	dcBmpButton.DeleteDC();
	ReleaseDC(&dcBmpButton);
	m_bitmap.DeleteObject();
	m_bitmap.Detach();
	
	// states

	// state focused
	if(state & ODS_FOCUS) {

		CPen pen;
		LOGPEN logPen;
		logPen.lopnColor = RGB(0,0,0);
		logPen.lopnStyle = PS_SOLID;
		logPen.lopnWidth = CPoint(2, 2);
		pen.CreatePenIndirect(&logPen);
		pDC->SelectObject(&pen);
		CBrush brushFocus;
		LOGBRUSH logBrush;
		logBrush.lbStyle = BS_NULL;		// transparent brush
		brushFocus.CreateBrushIndirect(&logBrush);
		pDC->SelectObject(&brushFocus);
		// marking around the ellipse
		pDC->Ellipse(0, 0, rect.Width(), rect.Height());
		brushFocus.DeleteObject();

		DrawFocus();
	}

	// delete the brush
	brush.DeleteObject();

	ReleaseDC(pDC);
}

/*****************************************************************************************/
/* the algorithm for the focus drawing                                                   */
/*****************************************************************************************/
void CEllipticButton::DrawFocus()
{
	CRect rect;
	GetClientRect(&rect);
	CDC* pDC = GetDC();;

	int x,y;
	int oldX = 0, oldY = 0;
	int i = 0;
	int step = rect.Width() + rect.Height();
	while (i < step) {
		x = (int)((rect.Width()/2 - 4) * sin(2*PI * i / step));
		y = (int)((rect.Height()/2 - 4) * cos(2*PI * i / step));
		if(((oldX - x) * (oldX - x) + (oldY - y) * (oldY - y)) > 3) {
			pDC->SetPixel(rect.Width()/2 + x, rect.Height()/2 + y, RGB(0,0,0));
			oldX = x;
			oldY = y;
		}
		i++;
	}
	CButton::SetFocus();
	ReleaseDC(pDC);
}

/*****************************************************************************************/
/* any stuff to improve the view of the button during mouse motion if the left mouse     */
/* button is pressed, for example if the mouse cross the border                          */
/*****************************************************************************************/
void CEllipticButton::OnMouseMove(UINT nFlags, CPoint point) 
{
	CRgn rgnEll, rgnAll, rgnRect;
	CRect rect;
	GetClientRect(rect);
	rgnRect.CreateRectRgnIndirect(rect);
	rgnEll.CreateEllipticRgnIndirect(rect);
	if(nFlags & MK_LBUTTON) {
		if(!rgnRect.PtInRegion(point)) {
			// if point is on the border - special handling
			if(rgnEll.PtInRegion(m_lastMousePos)) {
				m_selectedState = false;
				RedrawWindow();
			}
		}
		else {
			GetWindowRect(rect);
			if((nFlags & MK_LBUTTON)) {
				m_selectedState = true;		// it must display in selected state
			}
			if(!rgnEll.PtInRegion(point)) {
				m_selectedState = false;
			}
			if((rgnEll.PtInRegion(point) ^ rgnEll.PtInRegion(m_lastMousePos)) && (m_lastMousePos.x != INT_MAX) && (m_lastMousePos.y != INT_MAX)) {
				RedrawWindow();
			}
			CButton::OnMouseMove(nFlags, point);
		}
	}
	else {
		m_selectedState = false;
	}
	rgnRect.DeleteObject();
	rgnEll.DeleteObject();
	m_lastMousePos = point;
	return;
}

/*****************************************************************************************/
/* set the selected state member if the left mouse button is pressed                     */
/*****************************************************************************************/
void CEllipticButton::OnLButtonDown(UINT nFlags, CPoint point) 
{
	m_selectedState = true;

	CButton::OnLButtonDown(nFlags, point);
}

/*****************************************************************************************/
/* reset the selected state member if the left mouse button is pressed                   */
/* and prevent, that react OnLButtonUp outside of the ellipse                            */
/*****************************************************************************************/
void CEllipticButton::OnLButtonUp(UINT nFlags, CPoint point) 
{
	CRgn rgn;
	CRect rect;
	GetClientRect(rect);
	rgn.CreateEllipticRgnIndirect(rect);
	
	m_selectedState = false;
	if(!rgn.PtInRegion(point) && (point != CPoint(0, 0))) {
		SendMessage(WM_KILLFOCUS, 0,0);
		// kann ich hier die aktualisierung des Bildschirms verhindern?
		SendMessage(WM_SETFOCUS, 0,0);
	}
	else {
		CButton::OnLButtonDown(nFlags, point);
	}
}

/*****************************************************************************************/
/* initial settings                                                                      */
/*****************************************************************************************/
void CEllipticButton::PreSubclassWindow() 
{
	if(GetButtonStyle() & BS_DEFPUSHBUTTON) {
		m_defPushButton = true;
	}
	ModifyStyle(0, BS_OWNERDRAW | BS_DEFPUSHBUTTON);
	CButton::PreSubclassWindow();
}

/*****************************************************************************************/
/* handle ENTER key                                                                      */
/*****************************************************************************************/
BOOL CEllipticButton::PreTranslateMessage(MSG* pMsg) 
{
	if((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_RETURN)) {
		SendMessage(WM_LBUTTONDOWN, 0, 0);
		return TRUE;
	}
	if((pMsg->message == WM_KEYUP) && (pMsg->wParam == VK_RETURN)) {
		SendMessage(WM_LBUTTONUP, 0, 0);
		return TRUE;
	}
	
	return CButton::PreTranslateMessage(pMsg);
}
